/* 
 * Copyright (c) 2023 yann pascaud
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License along with this program.
 * If not, see https://www.gnu.org/licenses/
 */
#include <Arduino.h>
#include "NeoRammstein.h"

#ifdef USE_TINYNEOPIXEL_STATIC
#include <tinyNeoPixel_Static.h>
NeoRammstein::NeoRammstein(uint16_t n, int16_t pin, neoPixelType type, uint8_t *pxl) : tinyNeoPixel(n, pin, type, *&pxl)
{
  pinMode(pin, OUTPUT);
}

// Recopie de "Adafruit_NeoPixel" car absent de la lib "tinyNeoPixel_Static"
void NeoRammstein::fill(uint32_t c, uint16_t first, uint16_t count)
{
  uint16_t i, end;
  if (first >= this->numPixels())
  {
    return; // If first LED is past end of strip, nothing to do
  }

  // Calculate the index ONE AFTER the last pixel to fill
  if (count == 0)
  {
    // Fill to end of strip
    end = this->numPixels();
  }
  else
  {
    // Ensure that the loop won't go past the last pixel
    end = first + count;
    if (end > this->numPixels())
      end = this->numPixels();
  }

  for (i = first; i < end; i++)
  {
    this->setPixelColor(i, c);
  }
}

// Recopie de "Adafruit_NeoPixel" car absent de la lib "tinyNeoPixel_Static"
uint32_t NeoRammstein::ColorHSV(uint16_t hue, uint8_t sat, uint8_t val)
{

  uint8_t r, g, b;

  hue = (hue * 1530L + 32768) / 65536;

  if (hue < 510)
  { // Red to Green-1
    b = 0;
    if (hue < 255)
    { //   Red to Yellow-1
      r = 255;
      g = hue; //     g = 0 to 254
    }
    else
    {                //   Yellow to Green-1
      r = 510 - hue; //     r = 255 to 1
      g = 255;
    }
  }
  else if (hue < 1020)
  { // Green to Blue-1
    r = 0;
    if (hue < 765)
    { //   Green to Cyan-1
      g = 255;
      b = hue - 510; //     b = 0 to 254
    }
    else
    {                 //   Cyan to Blue-1
      g = 1020 - hue; //     g = 255 to 1
      b = 255;
    }
  }
  else if (hue < 1530)
  { // Blue to Red-1
    g = 0;
    if (hue < 1275)
    {                 //   Blue to Magenta-1
      r = hue - 1020; //     r = 0 to 254
      b = 255;
    }
    else
    { //   Magenta to Red-1
      r = 255;
      b = 1530 - hue; //     b = 255 to 1
    }
  }
  else
  { // Last 0.5 Red (quicker than % operator)
    r = 255;
    g = b = 0;
  }

  // Apply saturation and value to R,G,B, pack into 32-bit result:
  uint32_t v1 = 1 + val;  // 1 to 256; allows >>8 instead of /255
  uint16_t s1 = 1 + sat;  // 1 to 256; same reason
  uint8_t s2 = 255 - sat; // 255 to 0
  return ((((((r * s1) >> 8) + s2) * v1) & 0xff00) << 8) |
         (((((g * s1) >> 8) + s2) * v1) & 0xff00) |
         (((((b * s1) >> 8) + s2) * v1) >> 8);
}

#else
#include "src/Adafruit_NeoPixel/Adafruit_NeoPixel.h"
NeoRammstein::NeoRammstein(uint16_t n, int16_t pin, neoPixelType type) : Adafruit_NeoPixel(n, pin, type)
{
  this->begin();
}
#endif

///
/// Color blending
///
uint32_t NeoRammstein::substract(uint32_t color1, uint32_t color2)
{
  uint8_t r1, g1, b1;
  uint8_t r2, g2, b2;
  uint8_t r3, g3, b3;
  int16_t r, g, b;

  r1 = (uint8_t)(color1 >> 16),
  g1 = (uint8_t)(color1 >> 8),
  b1 = (uint8_t)(color1 >> 0);

  r2 = (uint8_t)(color2 >> 16),
  g2 = (uint8_t)(color2 >> 8),
  b2 = (uint8_t)(color2 >> 0);

  r = (int16_t)r1 - (int16_t)r2;
  g = (int16_t)g1 - (int16_t)g2;
  b = (int16_t)b1 - (int16_t)b2;
  if (r < 0)
    r = 0;
  if (g < 0)
    g = 0;
  if (b < 0)
    b = 0;

  return this->Color(r, g, b);
}

///
/// Color blending
///
uint32_t NeoRammstein::blend(uint32_t color1, uint32_t color2)
{
  uint8_t r1, g1, b1;
  uint8_t r2, g2, b2;
  uint8_t r3, g3, b3;

  r1 = (uint8_t)(color1 >> 16),
  g1 = (uint8_t)(color1 >> 8),
  b1 = (uint8_t)(color1 >> 0);

  r2 = (uint8_t)(color2 >> 16),
  g2 = (uint8_t)(color2 >> 8),
  b2 = (uint8_t)(color2 >> 0);

  return this->Color(constrain(r1 + r2, 0, 255), constrain(g1 + g2, 0, 255), constrain(b1 + b2, 0, 255));
}

///
/// Set color of LED
///
void NeoRammstein::addColor(uint8_t position, uint32_t color)
{
  uint32_t blended_color = this->blend(this->getPixelColor(position), color);
  this->setPixelColor(position, blended_color);
}

///
/// Set color of LED
///
void NeoRammstein::substractColor(uint8_t position, uint32_t color)
{
  uint32_t blended_color = this->substract(this->getPixelColor(position), color);
  this->setPixelColor(position, blended_color);
}

///
/// Set all colors
///
void NeoRammstein::startSequence()
{
  _elapsedTime = 0;
  _waitTime = 0;
}

void NeoRammstein::drawFire()
{
  if (this->delay())
  {
    uint8_t randomSize = 0;
    switch (this->getBrightness())
    {
    case 63:
      randomSize = 40;
      break;
    case 127:
      randomSize = 50;
      break;
    case 190:
      randomSize = 60;
      break;
    case 255:
      randomSize = 70;
      break;
    default:
      randomSize = 70;
      break;
    }
    this->fill(0);
    for (uint8_t i = 0; i < this->numPixels(); i++)
    {
      this->addColor(i, this->Color(MAIN_COLOR));
      int r = random(randomSize);
      uint32_t diff_color = this->Color(r * RANDOM_RED, r * RANDOM_GREEN, r * RANDOM_BLUE);
      this->substractColor(i, diff_color);
    }
    _waitTime = random(50, 150);
  }
}

bool NeoRammstein::delay()
{
  if (_elapsedTime == 0)
    _elapsedTime = millis();
  if (millis() - _elapsedTime > _waitTime)
  {
    _elapsedTime = 0;
    return true;
  }
  else
    return false;
}

void NeoRammstein::drawRainbow()
{
  if (this->delay())
  {
    this->fill(0);
    for (uint8_t i = CENTER_RING; i <= BACK_RING; i++)
    {
      this->setRingColor(i, ColorHSV(_hue + (uint16_t)i * 2000));
    }
    _hue = _hue + 50;
    _waitTime = 100;
  }
}

void NeoRammstein::setRingColor(uint8_t ring, uint32_t color)
{
  // There is 4 petals, organisation of each :
  // 0 : INNER_RING
  // 1-3 : MIDDLE_RING
  // 4-7 : OUTTER_RING
  // 8-10 : BACK_RING
  //
  // outside  petals :
  // 0 : CENTER_RING

  switch (ring)
  {
  case CENTER_RING:
    this->fill(color, 0, 1);
    break;

  case INNER_RING:
    for (size_t i = 0; i < LEAFS_NUM; i++)
    {
      this->fill(color, 1 + 0 + i * LEAFS_SIZE, 1);
    }
    break;

  case MIDDLE_RING:
    for (size_t i = 0; i < LEAFS_NUM; i++)
    {
      this->fill(color, 1 + 1 + i * LEAFS_SIZE, 3);
    }
    break;

  case OUTTER_RING:
    for (size_t i = 0; i < LEAFS_NUM; i++)
    {
      this->fill(color, 1 + 4 + i * LEAFS_SIZE, 4);
    }
    break;

  case BACK_RING:
    for (size_t i = 0; i < LEAFS_NUM; i++)
    {
      this->fill(color, 1 + 8 + i * LEAFS_SIZE, 3);
    }
    break;

  default:
    break;
  }
}
